% This program is related to the example of LDPC codes of chapter 8
% A slight modification is introduced in this program with respect to the
% way this example has been solved in Chapter 8
% Coefficients Q and R are calculated by the simplified method of 
% MacKay-Neal, so %that values R and Q that this program shows are the same 
% as those seen in tables 8.2 to 8.12, but normalised. 
% This means that for instance R12(0)= 0.0051 is normalised as
% R12N(0)=R12(0)/(R12(0)+R12(1))= 0.719 = 0.0051/(0.0051+0.0020)
% this program will show normalised values like R12N(0)=0.719

% Please note that q1 and q2 are set for reading the second received word of
% the file r_8.txt which corresponds to the case shown in the example of 
% chapter 8
% You can use this program for simulating other LDPC codes, by properly
% changing parameters N1, N2, LEN, NUMER_IT, but please remember that you
% need to create files H.txt and G.txt that correspond to the simulated
% case

% If files H.txt and G.txt are available, Inputword_q1, r_q1 
% can be generated using the program gen_words_ldpc_3.m

% Authors: Jorge Castieira and Patrick Farrell 
% 22/05/2006
% updated 03/11/2006

clear all;

N1=8; %This is n-k
N2=12;% This is the code length n
LEN=2;%number of runs in this simulation
for NUMER_IT=16:1:16 %NUMER_IT is the number of iterations of the sum-product algorithm
   
iter=0;

H=zeros(N1,N2);%parity check matrix is allocated in a matrix of size N1xN2, (n-k)xn
G=zeros(N2-N1,N2);%generator matrix is allocated in a matrix of size N1xN2, kxn
t=zeros(1,N2);
c=zeros(1,N2);


fid = fopen('H.txt','r');%the parity check matrix is read form the H.txt file
H=fscanf(fid,'%d ',[N1,N2]);
fclose(fid);
fid = fopen('Hmat.txt','w');%a test of the reading of H.txt
for i=1:N1
fprintf(fid,'%d ',H(i,:));
fprintf(fid,'\n');
end
fclose(fid);

fid = fopen('G.txt','r');%the generator matrix is read form the G.txt file
G=fscanf(fid,'%d ',[N2-N1,N2]);
fclose(fid);
fid = fopen('Gmat.txt','w');%a test of the reading of G.txt
for i=1:N2-N1
fprintf(fid,'%d ',G(i,:));
fprintf(fid,'\n');
end
fclose(fid);

HT=H';   %traspose of the parity check matrix H
   
for i=1:N1
onesfil(i)=nnz(H(i,:));%number of ones per row of the matrix H
end
for i=1:N2
onescol(i)=nnz(HT(i,:));%number of ones per column of the matrix H
end
[fil,colum]=find(HT);%row and column indexes of the non zero elements of HT

count=1;
for i=1:N1
   for j=1:onesfil(i)
   
   check_node(i,j)=fil(count);%check_node indicates the row indexes where  
   count=count+1;             %there are 1s
   end
end
[col,filas]=find(H);
count=1;
for i=1:N2
   for j=1:onescol(i)
   
   symbol_node(i,j)=col(count);%symbol_node indicates the column indexes 
   count=count+1;              %where there are 1s
   end
end

errors=zeros(1,10);%A list of errors for different values of EbNo
sig=zeros(1,10);%initialisation of sigma values
for q1=8:8
   
errors(1,q1)=0;
Inputword=zeros(1,N2-N1);%Defining a variable Inputword, which is an array    
                         %of input messages
arch='Inputword_';
i1=q1;
fid1 = fopen([arch num2str(i1)],'r');
Inputword1=fscanf(fid1,'%d ',[(N2-N1),LEN]);%Input messages are block of  
fclose(fid1);                               %binary information read from a 
                                            %file called Inputword_q1
arch='r_';
i1=q1;
fid2 = fopen([arch num2str(i1)],'r');
   r1=fscanf(fid2,'%f ',[N2,LEN]);
fclose(fid2);
%received bits affected by AWGN are block of numbers read from files of 
%the form r_q1

for q2=2:LEN %this loop decodes LEN blocks of length N2
      stop=0;
      iter=0;
   Sigma=0.1*q1;
   Inputword2=Inputword1(:,q2);
   Inputword=Inputword2'; %vector Inputword contains the input messages
  
   r2=r1(:,q2);
   r=r2';
%vector r contains the samples of the channel, bits in polar format (+-1) 
%that were affected by AWGN
f0=zeros(1,N2);
f1=zeros(1,N2);
f1=1./(1+exp(-2*r/Sigma^2));
f0=1-f1;
%estimations calculated as a function of the channel samples

count=1;
for i=1:N1
   for j=1:onesfil(i)
      Q0(i,check_node(i,j))=f0(check_node(i,j));%Initialization of coefficients Q
      Q1(i,check_node(i,j))=f1(check_node(i,j));
   count=count+1;   
   end
end

while (iter<=NUMER_IT)&(stop==0)
deltaQ_Mat=Q0-Q1;%calculation of the coefficients deltaQ
   
for i=1:N1
   for j=1:onesfil(i)
     deltaR_Mat(i,check_node(i,j))=1; 
     for k=1:onesfil(i)
        if (j~=k)
            deltaR_Mat(i,check_node(i,j))=deltaR_Mat(i,check_node(i,j))*deltaQ_Mat(i,check_node(i,k));
        end
     end
   end
end

%calculation of the coefficients deltaR


 R0=0.5*(1+deltaR_Mat);
 R1=0.5*(1-deltaR_Mat);
 R0(find(R0==0.5))=0;
 R1(find(R1==0.5))=0;
R0
R1
pause
%Calculation of coefficients R0 and R1, avoiding the appearance of NAN in
%Matlab
  
 for i=1:N1
       g=zeros(N1,max(onesfil));
       for j=1:onesfil(i)
       b0=1;
       b1=1;

	
       for p1=1:onescol(check_node(i,j))               
             if (check_node(i,j)~=0)&(i~=symbol_node(check_node(i,j),p1))
                b0=b0*R0(symbol_node(check_node(i,j),p1),check_node(i,j));
             end
          	 if (check_node(i,j)~=0)&(i~=symbol_node(check_node(i,j),p1))
                 b1=b1*R1(symbol_node(check_node(i,j),p1),check_node(i,j));
             end
       end
  %Normalization of coefficients
  
    b0=f0(check_node(i,j))*b0;
    b1=f1(check_node(i,j))*b1;
          
          if(b0==0)&(b1==0)
             Q0(i,check_node(i,j))=0;
             Q1(i,check_node(i,j))=0;
          else
             
          g(i,j)=1/(b0+b1);
          Q0(i,check_node(i,j))=b0*g(i,j);
          Q1(i,check_node(i,j))=b1*g(i,j);
          end
       end
 end
%The above rutine determines coefficients Q

d0=ones(1,N2);
d1=ones(1,N2);
for j=1:N2
   for i=1:onescol(j);
      d0(j)=d0(j)*R0(symbol_node(j,i),j);
      d1(j)=d1(j)*R1(symbol_node(j,i),j);
    
   end
   d0(j)=d0(j)*f0(j);
   d1(j)=d1(j)*f1(j);
end
%The above rutine determines estimations of bits in the current iteration

bitd=zeros(1,N2);

for j=1:N2
   if d0(j)>d1(j)
      bitd(1,j)=0;
   else
      bitd(1,j)=1;%A decision is taken at this iteration
   end
end

bitd %bitd is the estimation of the corresponding bit at the current iteration
pause

Sc=zeros(1,N1);
Sc=mod(bitd*HT,2);
syndro=sum(Sc);   %Calculation of the syndrome vector


if (syndro==0)         
   stop=1;%The iteratrion is halted if syndrome is equal to zero
end
iter
if (syndro~=0)&(iter==NUMER_IT)
   output=fix(sign(r(1,N1+1:N2))+0.5);%otherwise the decoding continues 
                                      %until performing NUMER_IT iterations
   iter=iter+1;
	
else
output=bitd(1,N1+1:N2);
iter=iter+1;
end

end       
q2 
errors(1,q1)=errors(1,q1)+sum(abs(Inputword-output))
sig(1,q1)=q1*0.1;
end

Pbe=errors./((N2-N1)*LEN)
arch='errors_';%a file containing accumulatively the number of errors 
i1=N2;         %for each case is stored
i2=N1;
i3=NUMER_IT;
i4=Sigma;
i5='.txt';
fid = fopen([arch num2str(i1) num2str(i2) num2str(i3) num2str(i4) num2str(i5)],'w');
      fprintf(fid,'%1.5f ',sig);
      fprintf(fid,'\n');
      fprintf(fid,'%1.5f ',errors);
      fprintf(fid,'\n');
      fprintf(fid,'%1.5f ',Pbe);
      fprintf(fid,'\n');
		fprintf(fid,'%d ',i3);
      fprintf(fid,'\n');

      fclose(fid);


end

end
   