% This program is related to the example of LDPC codes of chapter 8
%A slight modification is introduced in this program with respect to the
%way this example has been solved in Chapter 8
%Coefficients Q and R are calculated by the simplified method of MacKay, so
%that values R and Q that this program shows are the same as those seen in 
%tables 8.2 to 8.12, but normalised. 
%This means that for instance R12(0)= 0.0051 is normalised as
%R12N(0)=R12(0)/(R12(0)+R12(1))= 0.719 = 0.0051/(0.0051+0.0020)
%this program will show normalised values like R12N(0)=0.719
% Authors: Jorge Castieira and Patrick Farrell 
% 22/05/2006
% updated 03/11/2006

clear all;

N1=8; %This is n-k
N2=12;% This is the code length n
LEN=1;%number of runs in this simulation
for NUMER_IT=16:1:16 %NUMER_IT is the number of iterations of the sum-product algorithm
   
iter=0;

H=zeros(N1,N2);%parity check matrix is allocated in a matrix of size N1xN2, (n-k)xn
G=zeros(N2-N1,N2);%generator matrix is allocated in a matrix of size N1xN2, kxn
t=zeros(1,N2);
c=zeros(1,N2);


fid = fopen('H.txt','r');%the parity check matrix is read form the H.txt file
H=fscanf(fid,'%d ',[N1,N2]);
fclose(fid);
fid = fopen('Hmat.txt','w');
for i=1:N1
fprintf(fid,'%d ',H(i,:));
fprintf(fid,'\n');
end
fclose(fid);

fid = fopen('G.txt','r');%the generator matrix is read form the G.txt file
G=fscanf(fid,'%d ',[N2-N1,N2]);
fclose(fid);
fid = fopen('Gmat.txt','w');
for i=1:N2-N1
fprintf(fid,'%d ',G(i,:));
fprintf(fid,'\n');
end
fclose(fid);

HT=H';   %traspose of the parity check matrix
   
for i=1:N1
onesfil(i)=nnz(H(i,:));%number of ones per row of the matrix H
end
for i=1:N2
onescol(i)=nnz(HT(i,:));%number of ones per column of the matrix H
end
[fil,colum]=find(HT);%row and column indexes of the non zero elements of HT

count=1;
for i=1:N1
   for j=1:onesfil(i)
   
   check_node(i,j)=fil(count);%check_node indicates the row indexes where there are 1s 
   count=count+1;   
   end
end
[col,filas]=find(H);
count=1;
for i=1:N2
   for j=1:onescol(i)
   
   symbol_node(i,j)=col(count);%symbol_node indicates the column indexes where there are 1s
   count=count+1;   
   end
end

errors=zeros(1,10);%A list of errors for different values of EbNo
sig=zeros(1,10);
for q1=8:8
   
errors(1,q1)=0;
Inputword=zeros(1,N2-N1);%Defining a variable Inputword, which is an array of input messages   



Inputword1(:,1)=[1 0 0 0]';
%input word of the example of chapter 8
r1(:,1)=[1.312940 2.658411 0.741345 2.174506 0.598081 -0.832325 -0.396160 -1.758551 1.490525 0.408402 -0.928990 1.076456]';
%received bits affected by AWGN are block of numbers read from files of the
%form r_q1, this the received vector of the example of chapter 8

for q2=1:LEN %this loop decodes LEN blocks of length N2=12
      stop=0;
      iter=0;
   Sigma=0.1*q1;
   Inputword2=Inputword1(:,q2);
   Inputword=Inputword2'; %vector inputword contains the input messages
   %pause
   r2=r1(:,q2);
   r=r2';
%vector r contains the samples of the channel, bits in polar format (+-1) 
%that were affected by AWGN

f0=zeros(1,N2);
f1=zeros(1,N2);
f1=1./(1+exp(-2*r/Sigma^2));
f0=1-f1;
%estimations calculated as a function of the channel samples

count=1;
for i=1:N1
   for j=1:onesfil(i)
      Q0(i,check_node(i,j))=f0(check_node(i,j));%Initialization of coefficients Q
      Q1(i,check_node(i,j))=f1(check_node(i,j));
   count=count+1;   
   end
end
Q0
Q1
%Q0 and Q1 appear as normalised numbers with respect to those shown in
%tables 8.2 to 8.12
pause

while (iter<=NUMER_IT)&(stop==0)
deltaQ_Mat=Q0-Q1;%calculation of the coefficients deltaQ
   
for i=1:N1
   for j=1:onesfil(i)
     deltaR_Mat(i,check_node(i,j))=1; 
     for k=1:onesfil(i)
        if (k~=j)
            deltaR_Mat(i,check_node(i,j))=deltaR_Mat(i,check_node(i,j))*deltaQ_Mat(i,check_node(i,k));
        end
     end
     
     
   end
end
%The above loop makes the calculation of coefficients deltaR


 R0=0.5*(1+deltaR_Mat);
 R1=0.5*(1-deltaR_Mat);
 R0(find(R0==0.5))=0;
 R1(find(R1==0.5))=0;

 R0
 R1
 pause;%values R appear as normalised values with respect to 
 %those seen in tables 8.2 to 8.12
 %By normalising values of these tables we get R0 and R1 as shown by this
 %program
 %Calculation of coefficients R0 and R1, avoiding the appearance of NAN in
%Matlab
  
 for i=1:N1
       g=zeros(N1,max(onesfil));
       for j=1:onesfil(i)
       b0=1;
       b1=1;

	
       for p1=1:onescol(check_node(i,j))               
             if (check_node(i,j)~=0)&(i~=symbol_node(check_node(i,j),p1))
                b0=b0*R0(symbol_node(check_node(i,j),p1),check_node(i,j));
             end
          	 if (check_node(i,j)~=0)&(i~=symbol_node(check_node(i,j),p1))
                 b1=b1*R1(symbol_node(check_node(i,j),p1),check_node(i,j));
              end
       end
  %Normalization of coefficients
  
    b0=f0(check_node(i,j))*b0;
          b1=f1(check_node(i,j))*b1;
          
          if(b0==0)&(b1==0)
             Q0(i,check_node(i,j))=0;
             Q1(i,check_node(i,j))=0;
          else
             
          g(i,j)=1/(b0+b1);
          Q0(i,check_node(i,j))=b0*g(i,j);
          Q1(i,check_node(i,j))=b1*g(i,j);
          end
       end
 end
%The above rutine determines coefficients Q


d0=ones(1,N2);
d1=ones(1,N2);
for j=1:N2
   for i=1:onescol(j);
      d0(j)=d0(j)*R0(symbol_node(j,i),j);
      d1(j)=d1(j)*R1(symbol_node(j,i),j);
    
   end
   d0(j)=d0(j)*f0(j);
   d1(j)=d1(j)*f1(j);
end
%The above rutine determines estimations of bits in the current iteration

bitd=zeros(1,N2);

for j=1:N2
   if d0(j)>d1(j)
      bitd(1,j)=0;
   else
      bitd(1,j)=1;%A decision is taken at this iteration
   end
end

bitd %bitd is the estimation of the corresponding bit at the current iteration
pause

Sc=zeros(1,N1);
Sc=mod(bitd*HT,2);
syndro=sum(Sc);   %Calculation of the syndrome vector


if (syndro==0)         
   stop=1;%The iteration is halted if syndrome is equal to zero
end
iter
if (syndro~=0)&(iter==NUMER_IT)
   output=fix(sign(r(1,N1+1:N2))+0.5);%otherwise the decoding continues 
                                      %until performing NUMER_IT iterations
   iter=iter+1;
	
else
output=bitd(1,N1+1:N2);
iter=iter+1;
end

end       

errors(1,q1)=errors(1,q1)+sum(abs(Inputword-output))
end

Pbe=errors./((N2-N1)*LEN)


end
end
   